#!/usr/bin/perl -w
use strict;
use Getopt::Long;
use File::Basename;

###############################################################
##################### DOCUMENTATION ###########################
## Author: Yun Li
## Date: 
## Goal: 
my $location = $0;
my $title = " Ligate Haplotypes 0.0.4";

print "$title\n";
print "\t\@: $location\n";
print "\tby: Yun Li(yunli\@med.unc.edu)\n\n";

## Input Parameters
## 

## Input Files
##

## Output Files

## NOTES:
## (1) takes only two input files for now
## (2) takes gz or bz2 compressed files
## (3) Current algorithm:
##            12345678
##         ...AAAAAAAA
##         ...aaaaaaaa
##                  BBBBBBBB...
##                  bbbbbbbb...
##       i  . Consider only heterozygous sites, find the longest stretch of agreement between AB
##       ii . ................................, ............................................. Ab
##       iii. Ligate the one with the longer stretch from i and ii

##     Earelier algorithm (in version 0.0.1, details below) fails to consider that fact that 
##              haplotypes towards the ends (right end of A and a, and left end of B and b) are of less good quality. 
##     In the following scenario: 
##          A   ....111111111112222
##          a   ....222222222221111
##          B       222211111111111....
##          b       111122222222222....
##
##          current algorithm will ligate AB and ab; while earlier algorithm will ligate Ab (IBS=8: 4 at each end vs IBS(AB)=7 consecutive at the middle) and aB.
##

## (2) Version 0.01 ligated two pairs of haplotypes by comparing IBS: 
##       if IBS(A,B) > IBS(A,b),  ligate AB and bb
##       else,                    ligate Ab and aB
##            12345678
##         ...AAAAAAAA
##         ...aaaaaaaa
##                  BBBBBBBB...
##                  bbbbbbbb...
##     use all overlapping parts when comparing (pos 7 and 8 above)
##     use part (s.t. very end of each part will NOT be used) in ligation
##         e.g., using A/a up to 7 and 8- from B/b
##         default, ~halfway in the overlapping region
##  
################# END OF DOCUMENTATION ########################
###############################################################
my $starttime = time();

my $MIN_NUM_OPTS = 4;

if($#ARGV<($MIN_NUM_OPTS*2 - 1))
{
	&usage();
	die "Failed to parse parameters\n\n";
}

my %opts = ();

	# Default Optional Options
$opts{j} = "HALF";
$opts{o} = "ligateHaplotypes.out";

Getopt::Long::GetOptions(\%opts,qw(
	ah=s
	ap=s
	bh=s
	bp=s
	o=s
	j=i
)) || die "Failed to parse options\n\n";

if ( (!defined $opts{ah}) || (!defined $opts{ap}) || (!defined $opts{bh}) || (!defined $opts{bp}) ) 
{
	&usage();
	die "Failed to parse parameters\n\n";
}

&printPar();

my @aPosArray = ();
my %aPosHash = ();
my @bPosArray = ();
my %bPosHash = ();
my $aFirstOverlappingIndex;
my $aJunctionIndex; #last ligation index in a 
my $bJunctionIndex; #first ligation index in b

&main();

my $endtime = time();
my $lapsetime = $endtime - $starttime;
print "\nAnalysis took $lapsetime seconds\n\n";

# begin sub
sub usage
{
	print "\n";
	print "Usage:\n\t";
	print "-ah\t Part One Haplotype (Required) \n\t";
	print "-ap\t Part One Positions (Required) \n\t";
	print "-bh\t Part Two Haplotype (Required) \n\t";
	print "-bp\t Part Two Positions (Required) \n\t";
	print "-o\t Output Prefix (Optional) \n\t";
	print "\n";
}

sub printPar
{
	print "\n";
	print "Parameters in Effect:\n";
	print "\t Part One Haplotype \n\t\t-ah '$opts{ah}'\n";
	print "\t Part One Positions \n\t\t-ap '$opts{ap}'\n";
	print "\t Part Two Haplotype \n\t\t-bh '$opts{bh}'\n";
	print "\t Part Two Positions \n\t\t-bp '$opts{bp}'\n";
	print "\t Output Prefix \n\t\t-o '$opts{o}'\n";
	print "\n";
}

sub rm
{
	if(-e $_[0])
	{
		print "WARNING: $_[0] existed and removed!!\n\n";
		system "/bin/rm -rf $_[0]";
	}
}

sub bakrm
{
	if(-e $_[0])
	{
		my $tt = $_[0].".old";
		print "WARNING: $_[0] existed and moved to $tt !!\n\n";
		system "mv $_[0] $tt";
	}
}

sub swap
{
	my $tmp = $_[0];
	$_[0] = $_[1];
	$_[1] = $tmp;
}

sub swaparr
{
	my @tmp = $_[0];
	$_[0] = $_[1];
	$_[1] = @tmp;
}

#==================================================================
# Subroutine:
#   Open($filename, $handle)
#
# Usage:
#   Open('myfile.txt', \*IN);
#   Open($pedfile, \*PED);
#
#   Then refernce the handles <IN> and <PED> as usual.
#
# Description:
#   Open the file 'filename'. The file can be a normal file
#   or one compressed with gzip or bzip2.
#   Function does not return if the open fails.
#
# Arguments:
#   $filename - path to file to open
#   $handle - GLOB name of handle to use. Include '\*'
#
#==================================================================
sub Open {
    my ($filename, $handle) = @_;

    # Handle gzipped data
    if ($filename =~ /\.gz$/) {
        my $cmd = "gunzip -c $filename";
        open($handle, "$cmd |") ||
            die "Unable to invoke command '$cmd': $!\n";
        if (eof($handle)) { die "Cmd failed: '$cmd'\n"; }
        return 1;
    }
    # Handle bzip2 data
    if ($filename =~ /\.bz2$/) {
        my $cmd = "bunzip2 -c $filename";
        open($handle, "$cmd |") ||
            die "Unable to invoke command '$cmd': $!\n";
        if (eof($handle)) { die "Cmd failed: '$cmd'\n"; }
        return 2;
    }
    # Handle normal file
    open($handle, $filename) ||
        die "Unable to open file '$filename': $!\n";
    return 3;
}

# count number of lines: allow compressed files
sub Wcl
{
  my $nlines = 0;
  if(-e $_[0])
  {
    Open($_[0], \*IN) || die "cannot open file $_[0]\n\n";
    while (my $line = <IN>)
    {
      $nlines ++;
    }
    close(IN);
    return $nlines;
  }
}

sub main
{
	&loadPositions();
	&processHaplotypes();
}

sub loadPositions
{
	# read first set of positions
	Open($opts{ap}, \*IN) || die "cannot open file $opts{ap}\n\n";
	my $prevpos = 0;
	my $linenum = 0;
	while(my $line = <IN>)
	{
		$linenum++;
		chomp($line);
		my @linearr = (split(/\s+/, $line));
		if (scalar(@linearr)==0) 
		{
			next;
		}
		if (scalar(@linearr)!=1) 
		{
			print "ERROR: only one field is expected in $opts{ap}. Check line $linenum\n\n";
			die;
		}
		if ($line <= $prevpos)
		{
			print "ERROR: positions need to strictly increase in $opts{ap}. Check line $linenum\n\n";
			die;
		}
		push @aPosArray, $line;
		$aPosHash{$line} = $#aPosArray;
		$prevpos = $line;
	}
	close(IN);
	printf("Read %d positions from the first set\n", scalar(@aPosArray));
	
	# read second set of positions
	open(IN, $opts{bp}) || die "cannot open file $opts{bp}\n\n";
	$prevpos = 0;
	$linenum = 0;
	while(my $line = <IN>)
	{
		$linenum++;
		chomp($line);
		my @linearr = (split(/\s+/, $line));
		if (scalar(@linearr)==0) 
		{
			next;
		}
		if (scalar(@linearr)!=1) 
		{
			print "ERROR: only one field is expected in $opts{bp}. Check line $linenum\n\n";
			die;
		}
		if ($line <= $prevpos)
		{
			print "ERROR: positions need to strictly increase in $opts{bp}. Check line $linenum\n\n";
			die;
		}
		push @bPosArray, $line;
		$bPosHash{$line} = $#bPosArray;
		$prevpos = $line;
	}
	close(IN);
	printf("Read %d positions from the second set\n", scalar(@bPosArray));
	
	# sanity check (1) second set after first set
	if ($bPosArray[0] < $aPosArray[0] || $bPosArray[0] > $aPosArray[$#aPosArray])
	{
		print "ERROR: second set before first set or no overlapping\n\n";
		die;
	}
	
	# sanity check (2) identical set of positions within overlapping region
	if (!(exists $aPosHash{$bPosArray[0]}))
	{
		print "ERROR: first position in second set NOT in first set\n\n";
		die;
	}
	
	$aFirstOverlappingIndex = $aPosHash{$bPosArray[0]};
	my $aindex = $aFirstOverlappingIndex;
	for (my $i = 1; $i <= $#bPosArray; $i++)
	{
		# only check the overlapping region
		if ($bPosArray[$i] > $aPosArray[$#aPosArray])
		{
			last;
		}
		if (!(exists $aPosHash{$bPosArray[$i]}))
		{
			print "ERROR: position $bPosArray[$i] in second set NOT in first set\n\n";
			die;
		}
		my $aNewIndex = $aPosHash{$bPosArray[$i]};
		if ($aNewIndex != ($aindex + 1)) 
		{
			print "ERROR: not identical set of SNPs in overlapping region. Extra positions in the first set before $bPosArray[$i]\n\n";
			die;
		}
		$aindex = $aNewIndex;
	}
	
	# find the Half-way junction point: two indices (end index of A and start index of B)
	if ($opts{j} eq "HALF") # default
	{
		$aJunctionIndex = int(($aFirstOverlappingIndex + $#aPosArray)/2);
		$bJunctionIndex = $bPosHash{$aPosArray[$aJunctionIndex+1]};
	}

	print "Junction indices and positions are: $aJunctionIndex and $bJunctionIndex, $aPosArray[$aJunctionIndex] and $bPosArray[$bJunctionIndex]  (Half-way: NOT used for haplotype ligation)\n";
	#when ligating haplotypes, different people have different junctions
	
	my $opf = $opts{o}.".pos";
	&bakrm($opf);
	open(OUT, ">>$opf") || die "cannot create file $opf\n\n";
	for (my $i = 0; $i <= $aJunctionIndex; $i++)
	{
		print OUT "$aPosArray[$i]\n";
	}
	for (my $j = $bJunctionIndex; $j <= $#bPosArray; $j++)
	{
		print OUT "$bPosArray[$j]\n";
	}
	close(OUT);

}

sub processHaplotypes
{
	my $aNumLines = &Wcl($opts{ah});
	my $bNumLines = &Wcl($opts{bh});
	if ($aNumLines != $bNumLines)
	{
		print "ERROR: #lines differs in $opts{ah} and $opts{bh}\n\n";
		die;
	}
	
	Open($opts{ah}, \*ONE) || die "cannot open file $opts{ah}\n\n";
	Open($opts{bh}, \*TWO) || die "cannot open file $opts{bh}\n\n";
	my $ohf = $opts{o}.".hap";
	&bakrm($ohf);
	open(OUT, ">>$ohf") || die "cannot create file $ohf\n\n";
	while (my $linea1 = <ONE>)
	{
		my $linea2 = <ONE>;
		my $lineb1 = <TWO>;
		my $lineb2 = <TWO>;
		chomp($linea1);
		chomp($linea2);
		chomp($lineb1);
		chomp($lineb2);
		my @linearra1 = (split(/\s+/, $linea1));
		my @linearra2 = (split(/\s+/, $linea2));
		my @linearrb1 = (split(/\s+/, $lineb1));
		my @linearrb2 = (split(/\s+/, $lineb2));
		if ($#linearra1 != 2 || $#linearra2 != 2 || $#linearrb1 != 2 || $#linearrb2 != 2)
		{
			print "ERROR: number of elements in hap files has to be 3\n\n";
			die;
		}
		
		if ($linearra1[0] ne $linearra2[0] || $linearra1[0] ne $linearrb1[0] || $linearra1[0] ne $linearrb2[0])
		{
			print "ERROR: the four chromosomes read are NOT for a same individual $linearra1[0]\n\n";
			die;
		}
		
		my @hapa1 = (split(//, $linearra1[2]));
		my @hapa2 = (split(//, $linearra2[2]));
		my @hapb1 = (split(//, $linearrb1[2]));
		my @hapb2 = (split(//, $linearrb2[2]));
		if ($#hapa1 != $#aPosArray || $#hapa2 != $#aPosArray || $#hapb1 != $#bPosArray || $#hapb2 != $#bPosArray )
		{
			print "ERROR: # of alleles doesn't match position file(s) for  individual $linearra1[0]\n\n";
			die;
		}
		
		# find the longest stretch of agreement between hapa1(A) and hapb1(B)
		my $longestAgreementStrechLengthAB = 0;
		my $longestStretchStartAB = 0; 
		my $longestStretchEndAB = 0;
		my $agreementStrechLength = 0;
		my $stretchStart = 0;
		my $stretchEnd = 0;
		
		for (my $i = $aFirstOverlappingIndex, my $j = 0; $i <= $#aPosArray; $i++, $j++)
		{
			#if homozygous
			if ( (uc($hapa1[$i]) eq uc($hapa2[$i])) || 
			     (uc($hapb1[$j]) eq uc($hapb2[$j])) ) 
			{
				next;
			}
			
			#focus only at heterozygous sites
			if (uc($hapa1[$i]) eq uc($hapb1[$j])) 
			{
				if ($agreementStrechLength == 0)
				{
					$stretchStart = $j;
				}
				$agreementStrechLength ++;
				
			} # end of if agree
			else #not agree
			{
				if ($agreementStrechLength > $longestAgreementStrechLengthAB)
				{
					$longestStretchStartAB = $stretchStart;
					$longestStretchEndAB = $j - 1;
					$longestAgreementStrechLengthAB = $agreementStrechLength;
				}
				
				# reset length of agreement-stretch
				$agreementStrechLength = 0;
			} # end of else
		}
		# if last (checked <=> het) site agrees
		if ($agreementStrechLength > $longestAgreementStrechLengthAB)
		{
			$longestStretchStartAB = $stretchStart;
			$longestStretchEndAB = $#aPosArray - $aFirstOverlappingIndex;
			$longestAgreementStrechLengthAB = $agreementStrechLength;
		}
#print "$linearra1[0] longest stretch length $longestAgreementStrechLengthAB junction [$longestStretchStartAB, $longestStretchEndAB]\n";
		# find the longest stretch of agreement between hapa1(A) and hapb2(b)
		my $longestAgreementStrechLengthAb = 0;
		my $longestStretchStartAb = 0; 
		my $longestStretchEndAb = 0;
		$agreementStrechLength = 0;
		$stretchStart = 0;
		$stretchEnd = 0;
		
		for (my $i = $aFirstOverlappingIndex, my $j = 0; $i <= $#aPosArray; $i++, $j++)
		{
			#if homozygous
			if ( (uc($hapa1[$i]) eq uc($hapa2[$i])) || 
			     (uc($hapb1[$j]) eq uc($hapb2[$j])) ) 
			{
				next;
			}
			
			#focus only at heterozygous sites
			if (uc($hapa1[$i]) eq uc($hapb2[$j])) 
			{
				if ($agreementStrechLength == 0)
				{
					$stretchStart = $j;
				}
				$agreementStrechLength ++;
			} # end of if agree
			else #not agree
			{
				if ($agreementStrechLength > $longestAgreementStrechLengthAb)
				{
					$longestStretchStartAb = $stretchStart;
					$longestStretchEndAb = $j - 1;
					$longestAgreementStrechLengthAb = $agreementStrechLength;
				}
				
				# reset length of agreement-stretch
				$agreementStrechLength = 0;
			} # end of else
		}
		# if last (checked <=> het) site agrees
		if ($agreementStrechLength > $longestAgreementStrechLengthAb)
		{
			$longestStretchStartAb = $stretchStart;
			$longestStretchEndAb = $#aPosArray - $aFirstOverlappingIndex;
			$longestAgreementStrechLengthAb = $agreementStrechLength;
		}
#print "$linearra1[0] longest stretch length $longestAgreementStrechLengthAb junction [$longestStretchStartAb, $longestStretchEndAb]\n";
		
		my $bJunctionIndex = $longestStretchStartAB; #copy agreement-stretch from B
		# swap if Ab matches better than AB
		if ($longestStretchStartAb > $longestStretchStartAB)
		{
			$bJunctionIndex = $longestStretchStartAb;
			&swap($linearrb1[2], $linearrb2[2]);
			@hapb1 = (split(//, $linearrb1[2]));
			@hapb2 = (split(//, $linearrb2[2]));
		}
		
		$aJunctionIndex = $aFirstOverlappingIndex + $bJunctionIndex - 1;
		
		# output
		print OUT "$linearra1[0] HAPLO1 ";
		for (my $i = 0; $i <= $aJunctionIndex; $i++)
		{
			print OUT "$hapa1[$i]";
		}
		for (my $j = $bJunctionIndex; $j <= $#bPosArray; $j++)
		{
			print OUT "$hapb1[$j]";
		}
		print OUT "\n";
		print OUT "$linearra1[0] HAPLO2 ";
		for (my $i = 0; $i <= $aJunctionIndex; $i++)
		{
			print OUT "$hapa2[$i]";
		}
		for (my $j = $bJunctionIndex; $j <= $#bPosArray; $j++)
		{
			print OUT "$hapb2[$j]";
		}
		print OUT "\n";
	}
	close(ONE);
	close(TWO);
	close(OUT);
}

